@@ -2,13 +2,14 @@
 CmdStan arguments
 """
 import os
+import logging
 from time import time
 from enum import Enum, auto
 from numbers import Integral, Real
 from typing import List, Union
 from numpy.random import RandomState
 
-from cmdstanpy.utils import read_metric
+from cmdstanpy.utils import read_metric, get_logger
 
 
 class Method(Enum):
@@ -35,8 +36,11 @@ class SamplerArgs:
         max_treedepth: int = None,
         metric: Union[str, List[str]] = None,
         step_size: Union[float, List[float]] = None,
-        adapt_engaged: bool = None,
+        adapt_engaged: bool = True,
         adapt_delta: float = None,
+        adapt_init_phase: int = None,
+        adapt_metric_window: int = None,
+        adapt_step_size: int = None,
         fixed_param: bool = False,
     ) -> None:
         """Initialize object."""
@@ -50,6 +54,9 @@ class SamplerArgs:
         self.step_size = step_size
         self.adapt_engaged = adapt_engaged
         self.adapt_delta = adapt_delta
+        self.adapt_init_phase = adapt_init_phase
+        self.adapt_metric_window = adapt_metric_window
+        self.adapt_step_size = adapt_step_size
         self.fixed_param = fixed_param
         self.diagnostic_file = None
 
@@ -65,6 +72,30 @@ class SamplerArgs:
             raise ValueError(
                 'sampler expects number of chains to be greater than 0'
             )
+        if not (
+            self.adapt_delta is None
+            and self.adapt_init_phase is None
+            and self.adapt_metric_window is None
+            and self.adapt_step_size is None
+        ):
+            if self.adapt_engaged is False:
+                msg = 'conflicting arguments: adapt_engaged: False'
+                if self.adapt_delta is not None:
+                    msg = '{}, adapt_delta: {}'.format(msg, self.adapt_delta)
+                if self.adapt_init_phase is not None:
+                    msg = '{}, adapt_init_phase: {}'.format(
+                        msg, self.adapt_init_phase
+                    )
+                if self.adapt_metric_window is not None:
+                    msg = '{}, adapt_metric_window: {}'.format(
+                        msg, self.adapt_metric_window
+                    )
+                if self.adapt_step_size is not None:
+                    msg = '{}, adapt_step_size: {}'.format(
+                        msg, self.adapt_step_size
+                    )
+                raise ValueError(msg)
+
         if self.iter_warmup is not None:
             if self.iter_warmup < 0 or not isinstance(
                 self.iter_warmup, Integral
@@ -73,10 +104,9 @@ class SamplerArgs:
                     'iter_warmup must be a non-negative integer,'
                     ' found {}'.format(self.iter_warmup)
                 )
-            if self.adapt_engaged and self.iter_warmup == 0:
+            if self.iter_warmup > 0 and not self.adapt_engaged:
                 raise ValueError(
-                    'adaptation requested but 0 warmup iterations specified, '
-                    'must run warmup iterations'
+                    'adapt_engaged is False, cannot specify warmup iterations'
                 )
         if self.iter_sampling is not None:
             if self.iter_sampling < 0 or not isinstance(
@@ -89,7 +119,7 @@ class SamplerArgs:
         if self.thin is not None:
             if self.thin < 1 or not isinstance(self.thin, Integral):
                 raise ValueError(
-                    'thin must be a positive integer greater than 0,'
+                    'thin must be a positive integer,'
                     'found {}'.format(self.thin)
                 )
         if self.max_treedepth is not None:
@@ -97,14 +127,10 @@ class SamplerArgs:
                 self.max_treedepth, Integral
             ):
                 raise ValueError(
-                    'max_treedepth must be a positive integer greater than 0,'
+                    'max_treedepth must be a positive integer,'
                     ' found {}'.format(self.max_treedepth)
                 )
         if self.step_size is not None:
-            if self.fixed_param:
-                raise ValueError(
-                    'when fixed_param=True, cannot specify max_treedepth.'
-                )
             if isinstance(self.step_size, Real):
                 if self.step_size < 0:
                     raise ValueError(
@@ -113,8 +139,8 @@ class SamplerArgs:
             else:
                 if len(self.step_size) != chains:
                     raise ValueError(
-                        'number of step_sizes must match number of chains '
-                        ' found {} step_sizes for {} chains '.format(
+                        'number of step_sizes must match number of chains,'
+                        ' found {} step_sizes for {} chains'.format(
                             len(self.step_size), chains
                         )
                     )
@@ -137,8 +163,8 @@ class SamplerArgs:
             elif isinstance(self.metric, (list, tuple)):
                 if len(self.metric) != chains:
                     raise ValueError(
-                        'number of metric files must match number of chains '
-                        ' found {} metric files for {} chains '.format(
+                        'number of metric files must match number of chains,'
+                        ' found {} metric files for {} chains'.format(
                             len(self.metric), chains
                         )
                     )
@@ -162,8 +188,8 @@ class SamplerArgs:
                                     self.metric[0], metric
                                 )
                             )
-                        for j, dim in enumerate(dims):
-                            if dims[j] != dims2[j]:
+                        for dim, dim2 in zip(dims, dims2):
+                            if dim != dim2:
                                 raise ValueError(
                                     'metrics files {}, {},'
                                     ' inconsistent metrics'.format(
@@ -184,6 +210,30 @@ class SamplerArgs:
                     'adapt_delta must be between 0 and 1,'
                     ' found {}'.format(self.adapt_delta)
                 )
+        if self.adapt_init_phase is not None:
+            if self.adapt_init_phase < 0 or not isinstance(
+                self.adapt_init_phase, Integral
+            ):
+                raise ValueError(
+                    'adapt_init_phase must be a non-negative integer,'
+                    'found {}'.format(self.adapt_init_phase)
+                )
+        if self.adapt_metric_window is not None:
+            if self.adapt_metric_window < 0 or not isinstance(
+                self.adapt_metric_window, Integral
+            ):
+                raise ValueError(
+                    'adapt_metric_window must be a non-negative integer,'
+                    'found {}'.format(self.adapt_metric_window)
+                )
+        if self.adapt_step_size is not None:
+            if self.adapt_step_size < 0 or not isinstance(
+                self.adapt_step_size, Integral
+            ):
+                raise ValueError(
+                    'adapt_step_size must be a non-negative integer,'
+                    'found {}'.format(self.adapt_step_size)
+                )
 
         if self.fixed_param and (
             (self.iter_warmup is not None and self.iter_warmup > 0)
@@ -191,11 +241,16 @@ class SamplerArgs:
             or self.max_treedepth is not None
             or self.metric is not None
             or self.step_size is not None
-            or self.adapt_delta is not None
+            or not (
+                self.adapt_delta is None
+                and self.adapt_init_phase is None
+                and self.adapt_metric_window is None
+                and self.adapt_step_size is None
+            )
         ):
             raise ValueError(
-                'when fixed_param=True, cannot specify warmup or'
-                ' or any adaptation parameters.'
+                'when fixed_param=True, cannot specify warmup'
+                ' or adaptation parameters.'
             )
 
     def compose(self, idx: int, cmd: List) -> str:
@@ -231,15 +286,20 @@ class SamplerArgs:
                 cmd.append('metric_file={}'.format(self.metric_file))
             else:
                 cmd.append('metric_file={}'.format(self.metric_file[idx]))
-        if self.adapt_engaged is not None or self.adapt_delta is not None:
-            cmd.append('adapt')
-        if self.adapt_engaged is not None:
-            if self.adapt_engaged:
-                cmd.append('engaged=1')
-            else:
-                cmd.append('engaged=0')
+        cmd.append('adapt')
+        if self.adapt_engaged:
+            cmd.append('engaged=1')
+        else:
+            cmd.append('engaged=0')
         if self.adapt_delta is not None:
             cmd.append('delta={}'.format(self.adapt_delta))
+        if self.adapt_init_phase is not None:
+            cmd.append('init_buffer={}'.format(self.adapt_init_phase))
+        if self.adapt_metric_window is not None:
+            cmd.append('window={}'.format(self.adapt_metric_window))
+        if self.adapt_step_size is not None:
+            cmd.append('term_buffer={}'.format(self.adapt_step_size))
+
         return cmd
 
 
@@ -256,7 +316,7 @@ class OptimizeArgs:
         self.init_alpha = init_alpha
         self.iter = iter
 
-    def validate(self, chains=None) -> None:
+    def validate(self, chains=None) -> None:  # pylint: disable=unused-argument
         """
         Check arguments correctness and consistency.
         """
@@ -309,13 +369,13 @@ class GenerateQuantitiesArgs:
         """Initialize object."""
         self.sample_csv_files = csv_files
 
-    def validate(self, chains: int) -> None:
+    def validate(self, chains: int) -> None:  # pylint: disable=unused-argument
         """
         Check arguments correctness and consistency.
 
         * check that sample csv files exist
         """
-        for i, csv in enumerate(self.sample_csv_files):
+        for csv in self.sample_csv_files:
             if not os.path.exists(csv):
                 raise ValueError(
                     'Invalid path for sample csv file: {}'.format(csv)
@@ -357,7 +417,7 @@ class VariationalArgs:
         self.eval_elbo = eval_elbo
         self.output_samples = output_samples
 
-    def validate(self, chains=None) -> None:
+    def validate(self, chains=None) -> None:  # pylint: disable=unused-argument
         """
         Check arguments correctness and consistency.
         """
@@ -476,6 +536,7 @@ class CmdStanArgs:
         output_dir: str = None,
         save_diagnostics: bool = False,
         refresh: str = None,
+        logger: logging.Logger = None,
     ) -> None:
         """Initialize object."""
         self.model_name = model_name
@@ -497,6 +558,7 @@ class CmdStanArgs:
         elif isinstance(method_args, VariationalArgs):
             self.method = Method.VARIATIONAL
         self.method_args.validate(len(chain_ids) if chain_ids else None)
+        self._logger = logger or get_logger()
         self.validate()
 
     def validate(self) -> None:
@@ -519,25 +581,37 @@ class CmdStanArgs:
                     raise ValueError(
                         'invalid chain_id {}'.format(self.chain_ids[i])
                     )
-
         if self.output_dir is not None:
-            self.output_dir = os.path.realpath(os.path.expanduser(
-                self.output_dir))
-            if not os.path.exists(os.path.dirname(self.output_dir)):
+            self.output_dir = os.path.realpath(
+                os.path.expanduser(self.output_dir)
+            )
+            if not os.path.exists(self.output_dir):
+                try:
+                    os.makedirs(self.output_dir)
+                    self._logger.info(
+                        'created output directory: %s', self.output_dir
+                    )
+                except (RuntimeError, PermissionError):
+                    raise ValueError(
+                        'invalid path for output files, no such dir: {}'.format(
+                            self.output_dir
+                        )
+                    )
+            if not os.path.isdir(self.output_dir):
                 raise ValueError(
-                    'invalid path for output files, no such dir: {}'.format(
+                    'specified output_dir not a directory: {}'.format(
                         self.output_dir
                     )
                 )
             try:
                 testpath = os.path.join(self.output_dir, str(time()))
-                with open(testpath, 'w+') as fd:
+                with open(testpath, 'w+'):
                     pass
                 os.remove(testpath)  # cleanup
             except Exception:
                 raise ValueError(
-                    'invalid path for output files, '
-                    'cannot write to dir: {}'.format(self.output_dir)
+                    'invalid path for output files,'
+                    ' cannot write to dir: {}'.format(self.output_dir)
                 )
 
         if self.seed is None:
@@ -563,7 +637,7 @@ class CmdStanArgs:
 
                 if len(self.seed) != len(self.chain_ids):
                     raise ValueError(
-                        'number of seeds must match number of chains '
+                        'number of seeds must match number of chains,'
                         ' found {} seed for {} chains '.format(
                             len(self.seed), len(self.chain_ids)
                         )
@@ -602,7 +676,7 @@ class CmdStanArgs:
 
                 if len(self.inits) != len(self.chain_ids):
                     raise ValueError(
-                        'number of inits files must match number of chains '
+                        'number of inits files must match number of chains,'
                         ' found {} inits files for {} chains '.format(
                             len(self.inits), len(self.chain_ids)
                         )
@@ -619,8 +693,9 @@ class CmdStanArgs:
                             'no such file {}'.format(self.inits[i])
                         )
 
-    def compose_command(self, idx: int, csv_file: str,
-                        diagnostic_file: str = None) -> str:
+    def compose_command(
+        self, idx: int, csv_file: str, diagnostic_file: str = None
+    ) -> str:
         """
         Compose CmdStan command for non-default arguments.
         """
